=== modified file 'NEWS'
--- NEWS	2006-05-21 13:58:58 +0000
+++ NEWS	2006-05-22 14:53:12 +0000
@@ -2,32 +2,36 @@
 
   IMPROVEMENTS:
 
-   * On Unix, detect terminal width using an ioctl not just $COLUMNS.
-     Use terminal width for single-line logs from ``bzr log --line`` and
-     pending-merge display.  (Robert Widhopf-Fenk, Gustavo Niemeyer)
-     (Malone #3507)
+    * On Unix, detect terminal width using an ioctl not just $COLUMNS.
+      Use terminal width for single-line logs from ``bzr log --line`` and
+      pending-merge display.  (Robert Widhopf-Fenk, Gustavo Niemeyer)
+      (Malone #3507)
 
-   * On Windows, detect terminal width using GetConsoleScreenBufferInfo.
-     (Alexander Belchenko)
+    * On Windows, detect terminal widht using GetConsoleScreenBufferInfo.
+      (Alexander Belchenko)
   
-   * Show the correct number of revisions pushed when pushing a new branch.
-     (Robert Collins).
-
-   * 'bzr selftest' now shows a progress bar with the number of tests, and 
-     progress made. 'make check' shows tests in -v mode, to be more useful
-     for the PQM status window. (Robert Collins).
-
-   * 'bzr selftest --benchmark' will run a new benchmarking selftest.
-     (Robert Collins, Martin Pool).
-
-   * 'bzr commit' shows a progress bar. This is useful for commits over sftp
-     where commit can take an appreciable time. (Robert Collins)
-
-   * 'bzr add' is now less verbose in telling you what ignore globs were
-     matched by files being ignored. Instead it just tells you how many 
-     were ignored (because you might reasonably be expecting none to be
-     ignored). 'bzr add -v' is unchanged and will report every ignored
-     file. (Robert Collins).
+    * Show the correct number of revisions pushed when pushing a new branch.
+      (Robert Collins).
+
+    * 'bzr selftest' now shows a progress bar with the number of tests, and 
+      progress made. 'make check' shows tests in -v mode, to be more useful
+      for the PQM status window. (Robert Collins).
+
+    * 'bzr selftest --benchmark' will run a new benchmarking selftest.
+      (Robert Collins, Martin Pool).
+
+    * 'bzr commit' shows a progress bar. This is useful for commits over sftp
+      where commit can take an appreciable time. (Robert Collins)
+ 
+    * 'bzr add' is now less verbose in telling you what ignore globs were
+      matched by files being ignored. Instead it just tells you how many 
+      were ignored (because you might reasonably be expecting none to be
+      ignored). 'bzr add -v' is unchanged and will report every ignored
+      file. (Robert Collins).
+
+    * ftp now has a test server if medusa is installed. As part of testing,
+      ftp support has been improved, including support for supplying a
+      non-standard port. (John Arbash Meinel).
 
    * ``log --line`` shows the revision number.  (#5162, Alexander Belchenko)
 

=== modified file 'bzrlib/tests/test_transport_implementations.py'
--- bzrlib/tests/test_transport_implementations.py	2006-04-20 02:42:17 +0000
+++ bzrlib/tests/test_transport_implementations.py	2006-05-16 20:14:27 +0000
@@ -704,9 +704,10 @@
         except (ConnectionError, NoSuchFile), e:
             pass
         except (Exception), e:
-            self.failIf(True, 'Wrong exception thrown: %s' % e)
+            self.fail('Wrong exception thrown (%s): %s' 
+                        % (e.__class__.__name__, e))
         else:
-            self.failIf(True, 'Did not get the expected exception.')
+            self.fail('Did not get the expected ConnectionError or NoSuchFile.')
 
     def test_stat(self):
         # TODO: Test stat, just try once, and if it throws, stop testing

=== modified file 'bzrlib/transport/__init__.py'
--- bzrlib/transport/__init__.py	2006-04-12 04:45:32 +0000
+++ bzrlib/transport/__init__.py	2006-05-16 18:07:33 +0000
@@ -126,6 +126,35 @@
         urlparse.uses_netloc.append(protocol)
 
 
+def split_url(url):
+    if isinstance(url, unicode):
+        url = url.encode('utf-8')
+    (scheme, netloc, path, params,
+     query, fragment) = urlparse.urlparse(url, allow_fragments=False)
+    username = password = host = port = None
+    if '@' in netloc:
+        username, host = netloc.split('@', 1)
+        if ':' in username:
+            username, password = username.split(':', 1)
+            password = urllib.unquote(password)
+        username = urllib.unquote(username)
+    else:
+        host = netloc
+
+    if ':' in host:
+        host, port = host.rsplit(':', 1)
+        try:
+            port = int(port)
+        except ValueError:
+            # TODO: Should this be ConnectionError?
+            raise errors.TransportError('%s: invalid port number' % port)
+    host = urllib.unquote(host)
+
+    path = urllib.unquote(path)
+
+    return (scheme, username, password, host, port, path)
+
+
 class Transport(object):
     """This class encapsulates methods for retrieving or putting a file
     from/to a storage location.

=== modified file 'bzrlib/transport/ftp.py'
--- bzrlib/transport/ftp.py	2006-04-13 02:33:36 +0000
+++ bzrlib/transport/ftp.py	2006-05-22 05:32:51 +0000
@@ -25,38 +25,56 @@
 """
 
 from cStringIO import StringIO
+import asyncore
 import errno
 import ftplib
 import os
 import urllib
 import urlparse
 import stat
+import threading
 import time
 import random
 from warnings import warn
 
-
-from bzrlib.transport import Transport
-from bzrlib.errors import (TransportNotPossible, TransportError,
-                           NoSuchFile, FileExists, DirectoryNotEmpty)
+from bzrlib.transport import (
+    Transport,
+    Server,
+    split_url,
+    )
+import bzrlib.errors as errors
 from bzrlib.trace import mutter, warning
+import bzrlib.ui
+
+_have_medusa = False
+
+
+class FtpPathError(errors.PathError):
+    """FTP failed for path: %(path)s%(extra)s"""
 
 
 _FTP_cache = {}
-def _find_FTP(hostname, username, password, is_active):
+def _find_FTP(hostname, port, username, password, is_active):
     """Find an ftplib.FTP instance attached to this triplet."""
-    key = "%s|%s|%s|%s" % (hostname, username, password, is_active)
+    key = (hostname, port, username, password, is_active)
+    alt_key = (hostname, port, username, '********', is_active)
     if key not in _FTP_cache:
-        mutter("Constructing FTP instance against %r" % key)
-        _FTP_cache[key] = ftplib.FTP(hostname, username, password)
-        _FTP_cache[key].set_pasv(not is_active)
+        mutter("Constructing FTP instance against %r" % (alt_key,))
+        conn = ftplib.FTP()
+
+        conn.connect(host=hostname, port=port)
+        if username and username != 'anonymous' and not password:
+            password = bzrlib.ui.ui_factory.get_password(
+                prompt='FTP %(user)s@%(host)s password',
+                user=username, host=hostname)
+        conn.login(user=username, passwd=password)
+        conn.set_pasv(not is_active)
+
+        _FTP_cache[key] = conn
+
     return _FTP_cache[key]    
 
 
-class FtpTransportError(TransportError):
-    pass
-
-
 class FtpStatResult(object):
     def __init__(self, f, relpath):
         try:
@@ -80,36 +98,81 @@
     def __init__(self, base, _provided_instance=None):
         """Set the base path where files will be stored."""
         assert base.startswith('ftp://') or base.startswith('aftp://')
-        super(FtpTransport, self).__init__(base)
+
         self.is_active = base.startswith('aftp://')
         if self.is_active:
+            # urlparse won't handle aftp://
             base = base[1:]
-        (self._proto, self._host,
-            self._path, self._parameters,
-            self._query, self._fragment) = urlparse.urlparse(self.base)
+        if not base.endswith('/'):
+            base += '/'
+        (self._proto, self._username,
+            self._password, self._host,
+            self._port, self._path) = split_url(base)
+        base = self._unparse_url()
+
+        super(FtpTransport, self).__init__(base)
         self._FTP_instance = _provided_instance
 
+    def _unparse_url(self, path=None):
+        if path is None:
+            path = self._path
+        path = urllib.quote(path)
+        netloc = urllib.quote(self._host)
+        if self._username is not None:
+            netloc = '%s@%s' % (urllib.quote(self._username), netloc)
+        if self._port is not None:
+            netloc = '%s:%d' % (netloc, self._port)
+        return urlparse.urlunparse(('ftp', netloc, path, '', '', ''))
+
     def _get_FTP(self):
         """Return the ftplib.FTP instance for this object."""
         if self._FTP_instance is not None:
             return self._FTP_instance
         
         try:
-            username = ''
-            password = ''
-            hostname = self._host
-            if '@' in hostname:
-                username, hostname = hostname.split("@", 1)
-            if ':' in username:
-                username, password = username.split(":", 1)
-
-            self._FTP_instance = _find_FTP(hostname, username, password,
+            self._FTP_instance = _find_FTP(self._host, self._port,
+                                           self._username, self._password,
                                            self.is_active)
             return self._FTP_instance
         except ftplib.error_perm, e:
-            raise TransportError(msg="Error setting up connection: %s"
+            raise errors.TransportError(msg="Error setting up connection: %s"
                                     % str(e), orig_error=e)
 
+    def _translate_perm_error(self, err, path, extra=None, unknown_exc=FtpPathError):
+        """Try to translate an ftplib.error_perm exception.
+
+        :param err: The error to translate into a bzr error
+        :param path: The path which had problems
+        :param extra: Extra information which can be included
+        :param unknown_exc: If None, we will just raise the original exception
+                    otherwise we raise unknown_exc(path, extra=extra)
+        """
+        s = str(err).lower()
+        if not extra:
+            extra = str(err)
+        else:
+            extra += ': ' + str(err)
+        if ('no such file' in s
+            or 'could not open' in s
+            or 'no such dir' in s
+            ):
+            raise errors.NoSuchFile(path, extra=extra)
+        if ('file exists' in s):
+            raise errors.FileExists(path, extra=extra)
+        if ('not a directory' in s):
+            raise errors.PathError(path, extra=extra)
+
+        mutter('unable to understand error for path: %s: %s', path, err)
+
+        if unknown_exc:
+            raise unknown_exc(path, extra=extra)
+        # TODO: jam 20060516 Consider re-raising the error wrapped in 
+        #       something like TransportError, but this loses the traceback
+        #       Also, 'sftp' has a generic 'Failure' mode, which we use failure_exc
+        #       to handle. Consider doing something like that here.
+        #raise TransportError(msg='Error for path: %s' % (path,), orig_error=e)
+        raise
+
     def should_cache(self):
         """Return True if the data pulled across should be cached locally.
         """
@@ -127,11 +190,7 @@
     def _abspath(self, relpath):
         assert isinstance(relpath, basestring)
         relpath = urllib.unquote(relpath)
-        if isinstance(relpath, basestring):
-            relpath_parts = relpath.split('/')
-        else:
-            # TODO: Don't call this with an array - no magic interfaces
-            relpath_parts = relpath[:]
+        relpath_parts = relpath.split('/')
         if len(relpath_parts) > 1:
             if relpath_parts[0] == '':
                 raise ValueError("path %r within branch %r seems to be absolute"
@@ -153,30 +212,34 @@
         # Possibly, we could use urlparse.urljoin() here, but
         # I'm concerned about when it chooses to strip the last
         # portion of the path, and when it doesn't.
-        return '/'.join(basepath)
+        return '/'.join(basepath) or '/'
     
     def abspath(self, relpath):
         """Return the full url to the given relative path.
         This can be supplied with a string or a list
         """
         path = self._abspath(relpath)
-        return urlparse.urlunparse((self._proto,
-                self._host, path, '', '', ''))
+        return self._unparse_url(path)
 
     def has(self, relpath):
-        """Does the target location exist?
-
-        XXX: I assume we're never asked has(dirname) and thus I use
-        the FTP size command and assume that if it doesn't raise,
-        all is good.
-        """
+        """Does the target location exist?"""
+        # FIXME jam 20060516 We *do* ask about directories in the test suite
+        #       We don't seem to in the actual codebase
+        # XXX: I assume we're never asked has(dirname) and thus I use
+        # the FTP size command and assume that if it doesn't raise,
+        # all is good.
+        abspath = self._abspath(relpath)
         try:
             f = self._get_FTP()
-            s = f.size(self._abspath(relpath))
-            mutter("FTP has: %s" % self._abspath(relpath))
+            mutter('FTP has check: %s => %s', relpath, abspath)
+            s = f.size(abspath)
+            mutter("FTP has: %s", abspath)
             return True
-        except ftplib.error_perm:
-            mutter("FTP has not: %s" % self._abspath(relpath))
+        except ftplib.error_perm, e:
+            if ('is a directory' in str(e).lower()):
+                mutter("FTP has dir: %s: %s", abspath, e)
+                return True
+            mutter("FTP has not: %s: %s", abspath, e)
             return False
 
     def get(self, relpath, decode=False, retries=0):
@@ -191,26 +254,26 @@
         """
         # TODO: decode should be deprecated
         try:
-            mutter("FTP get: %s" % self._abspath(relpath))
+            mutter("FTP get: %s", self._abspath(relpath))
             f = self._get_FTP()
             ret = StringIO()
             f.retrbinary('RETR '+self._abspath(relpath), ret.write, 8192)
             ret.seek(0)
             return ret
         except ftplib.error_perm, e:
-            raise NoSuchFile(self.abspath(relpath), extra=str(e))
+            raise errors.NoSuchFile(self.abspath(relpath), extra=str(e))
         except ftplib.error_temp, e:
             if retries > _number_of_retries:
-                raise TransportError(msg="FTP temporary error during GET %s. Aborting."
+                raise errors.TransportError(msg="FTP temporary error during GET %s. Aborting."
                                      % self.abspath(relpath),
                                      orig_error=e)
             else:
-                warning("FTP temporary error: %s. Retrying." % str(e))
+                warning("FTP temporary error: %s. Retrying.", str(e))
                 self._FTP_instance = None
                 return self.get(relpath, decode, retries+1)
         except EOFError, e:
             if retries > _number_of_retries:
-                raise TransportError("FTP control connection closed during GET %s."
+                raise errors.TransportError("FTP control connection closed during GET %s."
                                      % self.abspath(relpath),
                                      orig_error=e)
             else:
@@ -229,42 +292,39 @@
 
         TODO: jam 20051215 ftp as a protocol seems to support chmod, but ftplib does not
         """
-        tmp_abspath = '%s.tmp.%.9f.%d.%d' % (self._abspath(relpath), time.time(),
+        abspath = self._abspath(relpath)
+        tmp_abspath = '%s.tmp.%.9f.%d.%d' % (abspath, time.time(),
                         os.getpid(), random.randint(0,0x7FFFFFFF))
         if not hasattr(fp, 'read'):
             fp = StringIO(fp)
         try:
-            mutter("FTP put: %s" % self._abspath(relpath))
+            mutter("FTP put: %s", abspath)
             f = self._get_FTP()
             try:
                 f.storbinary('STOR '+tmp_abspath, fp)
-                f.rename(tmp_abspath, self._abspath(relpath))
+                f.rename(tmp_abspath, abspath)
             except (ftplib.error_temp,EOFError), e:
                 warning("Failure during ftp PUT. Deleting temporary file.")
                 try:
                     f.delete(tmp_abspath)
                 except:
-                    warning("Failed to delete temporary file on the server.\nFile: %s"
-                            % tmp_abspath)
+                    warning("Failed to delete temporary file on the"
+                            " server.\nFile: %s", tmp_abspath)
                     raise e
                 raise
         except ftplib.error_perm, e:
-            if "no such file" in str(e).lower():
-                raise NoSuchFile("Error storing %s: %s"
-                                 % (self.abspath(relpath), str(e)), extra=e)
-            else:
-                raise FtpTransportError(orig_error=e)
+            self._translate_perm_error(e, abspath, extra='could not store')
         except ftplib.error_temp, e:
             if retries > _number_of_retries:
-                raise TransportError("FTP temporary error during PUT %s. Aborting."
+                raise errors.TransportError("FTP temporary error during PUT %s. Aborting."
                                      % self.abspath(relpath), orig_error=e)
             else:
-                warning("FTP temporary error: %s. Retrying." % str(e))
+                warning("FTP temporary error: %s. Retrying.", str(e))
                 self._FTP_instance = None
                 self.put(relpath, fp, mode, retries+1)
         except EOFError:
             if retries > _number_of_retries:
-                raise TransportError("FTP control connection closed during PUT %s."
+                raise errors.TransportError("FTP control connection closed during PUT %s."
                                      % self.abspath(relpath), orig_error=e)
             else:
                 warning("FTP control connection closed. Trying to reopen.")
@@ -272,70 +332,120 @@
                 self._FTP_instance = None
                 self.put(relpath, fp, mode, retries+1)
 
-
     def mkdir(self, relpath, mode=None):
         """Create a directory at the given path."""
+        abspath = self._abspath(relpath)
         try:
-            mutter("FTP mkd: %s" % self._abspath(relpath))
+            mutter("FTP mkd: %s", abspath)
             f = self._get_FTP()
-            try:
-                f.mkd(self._abspath(relpath))
-            except ftplib.error_perm, e:
-                s = str(e)
-                if 'File exists' in s:
-                    raise FileExists(self.abspath(relpath), extra=s)
-                else:
-                    raise
+            f.mkd(abspath)
         except ftplib.error_perm, e:
-            raise TransportError(orig_error=e)
+            self._translate_perm_error(e, abspath,
+                unknown_exc=errors.FileExists)
 
     def rmdir(self, rel_path):
         """Delete the directory at rel_path"""
+        abspath = self._abspath(rel_path)
         try:
-            mutter("FTP rmd: %s" % self._abspath(rel_path))
-
+            mutter("FTP rmd: %s", abspath)
             f = self._get_FTP()
-            f.rmd(self._abspath(rel_path))
+            f.rmd(abspath)
         except ftplib.error_perm, e:
-            if str(e).endswith("Directory not empty"):
-                raise DirectoryNotEmpty(self._abspath(rel_path), extra=str(e))
-            else:
-                raise TransportError(msg="Cannot remove directory at %s" % \
-                        self._abspath(rel_path), extra=str(e))
+            self._translate_perm_error(e, abspath, unknown_exc=errors.PathError)
 
-    def append(self, relpath, f):
+    def append(self, relpath, f, mode=None):
         """Append the text in the file-like object into the final
         location.
         """
-        raise TransportNotPossible('ftp does not support append()')
-
-    def copy(self, rel_from, rel_to):
-        """Copy the item at rel_from to the location at rel_to"""
-        raise TransportNotPossible('ftp does not (yet) support copy()')
+        abspath = self._abspath(relpath)
+        if self.has(relpath):
+            ftp = self._get_FTP()
+            result = ftp.size(abspath)
+        else:
+            result = 0
+
+        mutter("FTP appe to %s", abspath)
+        self._try_append(relpath, f.read(), mode)
+
+        return result
+
+    def _try_append(self, relpath, text, mode=None, retries=0):
+        """Try repeatedly to append the given text to the file at relpath.
+        
+        This is a recursive function. On errors, it will be called until the
+        number of retries is exceeded.
+        """
+        try:
+            abspath = self._abspath(relpath)
+            mutter("FTP appe (try %d) to %s", retries, abspath)
+            ftp = self._get_FTP()
+            ftp.voidcmd("TYPE I")
+            cmd = "APPE %s" % abspath
+            conn = ftp.transfercmd(cmd)
+            conn.sendall(text)
+            conn.close()
+            if mode:
+                self._setmode(relpath, mode)
+            ftp.getresp()
+        except ftplib.error_perm, e:
+            self._translate_perm_error(e, abspath, extra='error appending',
+                unknown_exc=errors.NoSuchFile)
+        except ftplib.error_temp, e:
+            if retries > _number_of_retries:
+                raise errors.TransportError("FTP temporary error during APPEND %s." \
+                        "Aborting." % abspath, orig_error=e)
+            else:
+                warning("FTP temporary error: %s. Retrying.", str(e))
+                self._FTP_instance = None
+                self._try_append(relpath, text, mode, retries+1)
+
+    def _setmode(self, relpath, mode):
+        """Set permissions on a path.
+
+        Only set permissions if the FTP server supports the 'SITE CHMOD'
+        extension.
+        """
+        try:
+            mutter("FTP site chmod: setting permissions to %s on %s",
+                str(mode), self._abspath(relpath))
+            ftp = self._get_FTP()
+            cmd = "SITE CHMOD %s %s" % (self._abspath(relpath), str(mode))
+            ftp.sendcmd(cmd)
+        except ftplib.error_perm, e:
+            # Command probably not available on this server
+            warning("FTP Could not set permissions to %s on %s. %s",
+                    str(mode), self._abspath(relpath), str(e))
+
+    # TODO: jam 20060516 I believe ftp allows you to tell an ftp server
+    #       to copy something to another machine. And you may be able
+    #       to give it its own address as the 'to' location.
+    #       So implement a fancier 'copy()'
 
     def move(self, rel_from, rel_to):
         """Move the item at rel_from to the location at rel_to"""
+        abs_from = self._abspath(rel_from)
+        abs_to = self._abspath(rel_to)
         try:
-            mutter("FTP mv: %s => %s" % (self._abspath(rel_from),
-                                         self._abspath(rel_to)))
+            mutter("FTP mv: %s => %s", abs_from, abs_to)
             f = self._get_FTP()
-            f.rename(self._abspath(rel_from), self._abspath(rel_to))
+            f.rename(abs_from, abs_to)
         except ftplib.error_perm, e:
-            raise TransportError(orig_error=e)
+            self._translate_perm_error(e, abs_from,
+                extra='unable to rename to %r' % (rel_to,), 
+                unknown_exc=errors.PathError)
 
     rename = move
 
     def delete(self, relpath):
         """Delete the item at relpath"""
+        abspath = self._abspath(relpath)
         try:
-            mutter("FTP rm: %s" % self._abspath(relpath))
+            mutter("FTP rm: %s", abspath)
             f = self._get_FTP()
-            f.delete(self._abspath(relpath))
+            f.delete(abspath)
         except ftplib.error_perm, e:
-            if str(e).endswith("No such file or directory"):
-                raise NoSuchFile(self._abspath(relpath), extra=str(e))
-            else:
-                raise TransportError(orig_error=e)
+            self._translate_perm_error(e, abspath, 'error deleting',
+                unknown_exc=errors.NoSuchFile)
 
     def listable(self):
         """See Transport.listable."""
@@ -344,16 +454,17 @@
     def list_dir(self, relpath):
         """See Transport.list_dir."""
         try:
-            mutter("FTP nlst: %s" % self._abspath(relpath))
+            mutter("FTP nlst: %s", self._abspath(relpath))
             f = self._get_FTP()
             basepath = self._abspath(relpath)
-            # FTP.nlst returns paths prefixed by relpath, strip 'em
-            the_list = f.nlst(basepath)
-            stripped = [path[len(basepath)+1:] for path in the_list]
+            paths = f.nlst(basepath)
+            # If FTP.nlst returns paths prefixed by relpath, strip 'em
+            if paths and paths[0].startswith(basepath):
+                paths = [path[len(basepath)+1:] for path in paths]
             # Remove . and .. if present, and return
-            return [path for path in stripped if path not in (".", "..")]
+            return [path for path in paths if path not in (".", "..")]
         except ftplib.error_perm, e:
-            raise TransportError(orig_error=e)
+            self._translate_perm_error(e, relpath, extra='error with list_dir')
 
     def iter_files_recursive(self):
         """See Transport.iter_files_recursive.
@@ -371,18 +482,14 @@
                 yield relpath
 
     def stat(self, relpath):
-        """Return the stat information for a file.
-        """
+        """Return the stat information for a file."""
+        abspath = self._abspath(relpath)
         try:
-            mutter("FTP stat: %s" % self._abspath(relpath))
+            mutter("FTP stat: %s", abspath)
             f = self._get_FTP()
-            return FtpStatResult(f, self._abspath(relpath))
+            return FtpStatResult(f, abspath)
         except ftplib.error_perm, e:
-            if "no such file" in str(e).lower():
-                raise NoSuchFile("Error storing %s: %s"
-                                 % (self.abspath(relpath), str(e)), extra=e)
-            else:
-                raise FtpTransportError(orig_error=e)
+            self._translate_perm_error(e, abspath, extra='error w/ stat')
 
     def lock_read(self, relpath):
         """Lock the given file for shared (read) access.
@@ -406,7 +513,207 @@
         return self.lock_read(relpath)
 
 
+class FtpServer(Server):
+    """Common code for SFTP server facilities."""
+
+    def __init__(self):
+        self._root = None
+        self._ftp_server = None
+        self._port = None
+        self._async_thread = None
+        # ftp server logs
+        self.logs = []
+
+    def get_url(self):
+        """Calculate an ftp url to this server."""
+        return 'ftp://foo:bar@localhost:%d/' % (self._port)
+
+#    def get_bogus_url(self):
+#        """Return a URL which cannot be connected to."""
+#        return 'ftp://127.0.0.1:1'
+
+    def log(self, message):
+        """This is used by medusa.ftp_server to log connections, etc."""
+        self.logs.append(message)
+
+    def setUp(self):
+
+        if not _have_medusa:
+            raise RuntimeError('Must have medusa to run the FtpServer')
+
+        self._root = os.getcwdu()
+        self._ftp_server = _ftp_server(
+            authorizer=_test_authorizer(root=self._root),
+            ip='localhost',
+            port=0, # bind to a random port
+            resolver=None,
+            logger_object=self # Use FtpServer.log() for messages
+            )
+        self._port = self._ftp_server.getsockname()[1]
+        # Don't let it loop forever, or handle an infinite number of requests.
+        # In this case it will run for 100s, or 1000 requests
+        self._async_thread = threading.Thread(target=asyncore.loop,
+                kwargs={'timeout':0.1, 'count':1000})
+        self._async_thread.setDaemon(True)
+        self._async_thread.start()
+
+    def tearDown(self):
+        """See bzrlib.transport.Server.tearDown."""
+        # have asyncore release the channel
+        self._ftp_server.del_channel()
+        asyncore.close_all()
+        self._async_thread.join()
+
+
+_ftp_channel = None
+_ftp_server = None
+_test_authorizer = None
+
+
+def _setup_medusa():
+    global _have_medusa, _ftp_channel, _ftp_server, _test_authorizer
+    try:
+        import medusa
+        import medusa.filesys
+        import medusa.ftp_server
+    except ImportError:
+        return False
+
+    _have_medusa = True
+
+    class test_authorizer(object):
+        """A custom Authorizer object for running the test suite.
+
+        The reason we cannot use dummy_authorizer, is because it sets the
+        channel to readonly, which we don't always want to do.
+        """
+
+        def __init__(self, root):
+            self.root = root
+
+        def authorize(self, channel, username, password):
+            """Return (success, reply_string, filesystem)"""
+            if not _have_medusa:
+                return 0, 'No Medusa.', None
+
+            channel.persona = -1, -1
+            if username == 'anonymous':
+                channel.read_only = 1
+            else:
+                channel.read_only = 0
+
+            return 1, 'OK.', medusa.filesys.os_filesystem(self.root)
+
+
+    class ftp_channel(medusa.ftp_server.ftp_channel):
+        """Customized ftp channel"""
+
+        def log(self, message):
+            """Redirect logging requests."""
+            mutter('_ftp_channel: %s', message)
+            
+        def log_info(self, message, type='info'):
+            """Redirect logging requests."""
+            mutter('_ftp_channel %s: %s', type, message)
+            
+        def cmd_rnfr(self, line):
+            """Prepare for renaming a file."""
+            self._renaming = line[1]
+            self.respond('350 Ready for RNTO')
+            # TODO: jam 20060516 in testing, the ftp server seems to
+            #       check that the file already exists, or it sends
+            #       550 RNFR command failed
+
+        def cmd_rnto(self, line):
+            """Rename a file based on the target given.
+
+            rnto must be called after calling rnfr.
+            """
+            if not self._renaming:
+                self.respond('503 RNFR required first.')
+            pfrom = self.filesystem.translate(self._renaming)
+            self._renaming = None
+            pto = self.filesystem.translate(line[1])
+            try:
+                os.rename(pfrom, pto)
+            except (IOError, OSError), e:
+                # TODO: jam 20060516 return custom responses based on
+                #       why the command failed
+                self.respond('550 RNTO failed: %s' % (e,))
+            except:
+                self.respond('550 RNTO failed')
+                # For a test server, we will go ahead and just die
+                raise
+            else:
+                self.respond('250 Rename successful.')
+
+        def cmd_size(self, line):
+            """Return the size of a file
+
+            This is overloaded to help the test suite determine if the 
+            target is a directory.
+            """
+            filename = line[1]
+            if not self.filesystem.isfile(filename):
+                if self.filesystem.isdir(filename):
+                    self.respond('550 "%s" is a directory' % (filename,))
+                else:
+                    self.respond('550 "%s" is not a file' % (filename,))
+            else:
+                self.respond('213 %d' 
+                    % (self.filesystem.stat(filename)[stat.ST_SIZE]),)
+
+        def cmd_mkd(self, line):
+            """Create a directory.
+
+            Overloaded because default implementation does not distinguish
+            *why* it cannot make a directory.
+            """
+            if len (line) != 2:
+                self.command_not_understood (string.join (line))
+            else:
+                path = line[1]
+                try:
+                    self.filesystem.mkdir (path)
+                    self.respond ('257 MKD command successful.')
+                except (IOError, OSError), e:
+                    self.respond ('550 error creating directory: %s' % (e,))
+                except:
+                    self.respond ('550 error creating directory.')
+
+
+    class ftp_server(medusa.ftp_server.ftp_server):
+        """Customize the behavior of the Medusa ftp_server.
+
+        There are a few warts on the ftp_server, based on how it expects
+        to be used.
+        """
+        _renaming = None
+        ftp_channel_class = ftp_channel
+
+        def __init__(self, *args, **kwargs):
+            mutter('Initializing _ftp_server: %r, %r', args, kwargs)
+            medusa.ftp_server.ftp_server.__init__(self, *args, **kwargs)
+
+        def log(self, message):
+            """Redirect logging requests."""
+            mutter('_ftp_server: %s', message)
+
+        def log_info(self, message, type='info'):
+            """Override the asyncore.log_info so we don't stipple the screen."""
+            mutter('_ftp_server %s: %s', type, message)
+
+    _test_authorizer = test_authorizer
+    _ftp_channel = ftp_channel
+    _ftp_server = ftp_server
+
+    return True
+
+
 def get_test_permutations():
     """Return the permutations to be used in testing."""
-    warn("There are no FTP transport provider tests yet.")
-    return []
+    if not _setup_medusa():
+        warn("You must install medusa (http://www.amk.ca/python/code/medusa.html) for FTP tests")
+        return []
+    else:
+        return [(FtpTransport, FtpServer)]

=== modified file 'bzrlib/transport/sftp.py'
--- bzrlib/transport/sftp.py	2006-05-18 03:47:41 +0000
+++ bzrlib/transport/sftp.py	2006-05-19 00:35:14 +0000
@@ -44,6 +44,7 @@
 from bzrlib.transport import (
     register_urlparse_netloc_protocol,
     Server,
+    split_url,
     Transport,
     urlescape,
     )
@@ -662,31 +663,8 @@
         return urlparse.urlunparse(('sftp', netloc, path, '', '', ''))
 
     def _split_url(self, url):
-        if isinstance(url, unicode):
-            url = url.encode('utf-8')
-        (scheme, netloc, path, params,
-         query, fragment) = urlparse.urlparse(url, allow_fragments=False)
+        (scheme, username, password, host, port, path) = split_url(url)
         assert scheme == 'sftp'
-        username = password = host = port = None
-        if '@' in netloc:
-            username, host = netloc.split('@', 1)
-            if ':' in username:
-                username, password = username.split(':', 1)
-                password = urllib.unquote(password)
-            username = urllib.unquote(username)
-        else:
-            host = netloc
-
-        if ':' in host:
-            host, port = host.rsplit(':', 1)
-            try:
-                port = int(port)
-            except ValueError:
-                # TODO: Should this be ConnectionError?
-                raise TransportError('%s: invalid port number' % port)
-        host = urllib.unquote(host)
-
-        path = urllib.unquote(path)
 
         # the initial slash should be removed from the path, and treated
         # as a homedir relative path (the path begins with a double slash

